package eu.jonahbauer.wizard.client.libgdx.screens;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.graphics.Color;
import com.badlogic.gdx.graphics.g2d.TextureAtlas;
import com.badlogic.gdx.math.Vector2;
import com.badlogic.gdx.scenes.scene2d.Action;
import com.badlogic.gdx.scenes.scene2d.Touchable;
import com.badlogic.gdx.scenes.scene2d.actions.MoveToAction;
import com.badlogic.gdx.scenes.scene2d.ui.Container;
import com.badlogic.gdx.scenes.scene2d.ui.Label;
import com.badlogic.gdx.scenes.scene2d.ui.VerticalGroup;
import com.badlogic.gdx.scenes.scene2d.utils.TextureRegionDrawable;
import com.badlogic.gdx.utils.Align;
import com.badlogic.gdx.utils.I18NBundle;
import eu.jonahbauer.wizard.client.libgdx.AnimationTimings;
import eu.jonahbauer.wizard.client.libgdx.GameAtlas;
import eu.jonahbauer.wizard.client.libgdx.WizardGame;
import eu.jonahbauer.wizard.client.libgdx.actors.game.CardActor;
import eu.jonahbauer.wizard.client.libgdx.actors.game.CardStack;
import eu.jonahbauer.wizard.client.libgdx.actors.game.CardsGroup;
import eu.jonahbauer.wizard.client.libgdx.actors.game.PadOfTruth;
import eu.jonahbauer.wizard.client.libgdx.actors.game.overlay.*;
import eu.jonahbauer.wizard.client.libgdx.state.Game;
import eu.jonahbauer.wizard.common.messages.observer.UserInputMessage;
import eu.jonahbauer.wizard.common.model.Card;
import lombok.Getter;
import org.jetbrains.annotations.Nls;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.*;
import java.util.concurrent.atomic.AtomicBoolean;

import static eu.jonahbauer.wizard.client.libgdx.actions.MyActions.*;

public class GameScreen extends MenuScreen {
    @Getter
    private TextureAtlas atlas;

    private Label.LabelStyle labelStyleDefault;
    private Label.LabelStyle labelStyleActive;

    private final UUID self;
    private final LinkedHashMap<UUID, String> players;
    private final List<UUID> orderedPlayers;

    private CardsGroup handCards;
    private CardStack cardStack;
    private PadOfTruth padOfTruth;

    @Getter
    private CardActor trumpCardActor;
    @Getter
    private CardActor trumpSuitActor;

    private VerticalGroup messages;
    private Label persistentMessage;

    private Action currentAction;
    private final Queue<Action> pendingActions = new LinkedList<>();

    private final Map<UUID, Seat> seats = new HashMap<>();
    private final Map<UUID, Label> nameLabels = new HashMap<>();

    private final AtomicBoolean pendingSync = new AtomicBoolean();

    public GameScreen(WizardGame game, @NotNull UUID self, @NotNull LinkedHashMap<UUID, String> players) {
        super(game);
        this.self = self;
        this.players = players;
        this.orderedPlayers = players.keySet().stream().toList();
    }

    //<editor-fold desc="Layout">
    @Override
    public void show() {
        super.show();
        atlas = new TextureAtlas(Gdx.files.internal(GameAtlas.$PATH));
//        overlay = new TextureRegionDrawable(game.data.uiskinAtlas.findRegion(UiskinAtlas.WHITE)).tint(new Color(0,0,0,0.5f));
        labelStyleDefault = game.data.skin.get(Label.LabelStyle.class);
        labelStyleActive = new Label.LabelStyle(labelStyleDefault);
        labelStyleActive.fontColor = Color.RED;

        seat();

        handCards = new CardsGroup(Collections.emptyList(), atlas);
        handCards.setOnClickListener(actor -> onCardClicked(actor.getCard()));
        var container = new Container<>(handCards);
        container.setPosition(360, 75);
        container.setSize(1200, CardActor.PREF_HEIGHT);
        container.fillX();

        messages = new VerticalGroup().columnCenter().bottom();
        messages.setPosition(360, 85 + CardActor.PREF_HEIGHT);
        messages.setSize(1200, 0);
        messages.setTouchable(Touchable.disabled);

        padOfTruth = new PadOfTruth(game.data.skin, new TextureRegionDrawable(atlas.findRegion(GameAtlas.PAD_OF_TRUTH)));
        padOfTruth.setPosition(WizardGame.WIDTH - 10 - PadOfTruth.EXTENDED_WIDTH, 10);
        padOfTruth.setOrigin(PadOfTruth.EXTENDED_WIDTH, 0);

        cardStack = new CardStack();
        cardStack.setHoverBounds(0.5f * WizardGame.WIDTH - 50, 0.5f * WizardGame.HEIGHT - 50, 100, 100);

        addLabels();

        Gdx.input.setInputProcessor(game.data.stage);
        game.data.stage.addActor(container);
        game.data.stage.addActor(cardStack);
        game.data.stage.addActor(padOfTruth);
        game.data.stage.addActor(messages);
    }

    private void seat() {
        var count = players.size();
        Seat[] seats = switch (count) {
            case 3 -> new Seat[] {Seat.BOTTOM, Seat.TOP_LEFT, Seat.TOP_RIGHT};
            case 4 -> new Seat[] {Seat.BOTTOM, Seat.LEFT, Seat.TOP, Seat.RIGHT};
            case 5 -> new Seat[] {Seat.BOTTOM, Seat.LEFT, Seat.TOP_LEFT, Seat.TOP_RIGHT, Seat.RIGHT};
            case 6 -> new Seat[] {Seat.BOTTOM, Seat.LEFT, Seat.TOP_LEFT, Seat.TOP, Seat.TOP_RIGHT, Seat.RIGHT};
            default -> throw new AssertionError();
        };

        int index = orderedPlayers.indexOf(self);
        for (int i = 0; i < count; i++) {
            var player = orderedPlayers.get((index + i) % count);
            var seat = seats[i];
            this.seats.put(player, seat);
        }
    }

    private void addLabels() {
        int i = 0;
        for (var entry : players.entrySet()) {
            UUID uuid = entry.getKey();
            String name = entry.getValue();

            padOfTruth.setName(i++, name);

            if (isSelf(uuid)) continue;
            var label = new Label("", game.data.skin);
            var seat = seats.get(uuid);
            label.setX(seat.getLabelX());
            label.setY(seat.getLabelY());
            label.setAlignment(seat.getLabelAlign());
            label.setText(name);
            nameLabels.put(uuid, label);
            game.data.stage.addActor(label);
        }
    }
    //</editor-fold>

    @Override
    protected void renderBackground(float delta) {
        float scale = Math.max(
                game.data.extendViewport.getWorldWidth() / WizardGame.WIDTH,
                game.data.extendViewport.getWorldHeight() / WizardGame.HEIGHT
        );
        game.batch.draw(game.data.background, 0,0, scale * WizardGame.WIDTH, scale * WizardGame.HEIGHT);
        game.batch.setColor(1, 1, 1, 0.25f);
        game.batch.draw(
                game.data.title,
                (game.data.extendViewport.getWorldWidth() - game.data.title.getRegionWidth() * 0.75f) / 2f,
                (game.data.extendViewport.getWorldHeight() - game.data.title.getRegionHeight() * 0.75f) / 2f,
                0.75f * game.data.title.getRegionWidth(),
                0.75f * game.data.title.getRegionHeight()
        );
    }

    @Override
    protected void renderForeground(float delta) {
        if (pendingActions.size() > 0) {
            var actions = game.data.stage.getRoot().getActions();
            boolean running = false;
            if (currentAction != null) {
                for (int i = 0; i < actions.size; i++) {
                    if (actions.get(i) == currentAction) {
                        running = true;
                        break;
                    }
                }
            }

            if (!running) {
                currentAction = pendingActions.poll();
                game.data.stage.addAction(currentAction);
            }
        } else if (pendingSync.getAndSet(false)) {
            game.getClient().execute(Game.class, Game::sync);
        }
    }

    private void execute(Action action) {
        pendingActions.add(action);
    }

    private void execute(Runnable runnable) {
        pendingActions.add(run(runnable));
    }

    private boolean isSelf(UUID uuid) {
        return self.equals(uuid);
    }

    public void onCardClicked(@NotNull Card card) {
        game.getClient().execute(Game.class, (s, c) -> s.onCardClicked(c, card));
    }

    public void onSuitClicked(@NotNull Card.Suit suit) {
        game.getClient().execute(Game.class, (s, c) -> s.onSuitClicked(c, suit));
    }

    public void onPredictionMade(int prediction) {
        game.getClient().execute(Game.class, (s, c) -> s.onPredictionMade(c, prediction));
    }

    /**
     * Resets all round-scoped and shows an {@linkplain StartRoundOverlay overlay}.
     */
    public void startRound(int round) {
        execute(parallel(
                run(() -> {
                    if (trumpCardActor != null) {
                        trumpCardActor.remove();
                    }

                    if (trumpSuitActor != null) {
                        trumpSuitActor.remove();
                    }

                    handCards.clearChildren();
                    cardStack.clearChildren();
                }),
                new StartRoundOverlay(this, round)
        ));
    }

    public void startTrick() {
        clearActivePlayer();
        execute(() -> cardStack.clearChildren());
    }

    public void finishTrick(@NotNull UUID player) {
        var seat = seats.getOrDefault(player, Seat.FALLBACK);

        var action = parallel();
        execute(sequence(
                run(() -> cardStack.removeAll().forEach(card -> {
                    var angle = (card.getRotation() % 360 + 360) % 360;
                    var rotation = rotateTo(angle < 90 || angle > 270 ? 0 : 180, AnimationTimings.STACK_FINISH_ROTATE);
                    rotation.setUseShortestDirection(true);

                    action.addAction(sequence(
                            targeting(card, changeParent(game.data.stage.getRoot())),
                            parallel(
                                    targeting(card, rotation),
                                    targeting(card, moveTo(
                                            seat.getFrontX() - card.getWidth() / 2,
                                            seat.getFrontY() - card.getHeight() / 2,
                                            AnimationTimings.STACK_FINISH_MOVE
                                    ))
                            ),
                            targeting(card, alpha(0, AnimationTimings.STACK_FINISH_FADE)),
                            removeActor(card)
                    ));
                })),
                action
        ));
    }

    /**
     * Updates the given players hand cards.
     */
    public void setHand(@NotNull UUID player, @NotNull List<Card> cards, boolean juggle) {
        if (isSelf(player)) {
            var sequence = sequence();
            sequence.addAction(run(() -> {
                var changes = handCards.update(cards);

                // animate card changes
                if (juggle) {
                    sequence.addAction(animateJuggle(changes.first(), changes.second()));
                }
            }));
            execute(sequence);
        }
    }

    /**
     * Adds the prediction for the given round and player to the {@linkplain #padOfTruth pad of truth} and
     * shows a corresponding message.
     */
    public void addPrediction(int round, @NotNull UUID player, int prediction, boolean changed) {
        if (isSelf(player)) {
            addMessage(game.messages.format("game.action." + (changed ? "change" : "make") + "_prediction.self", prediction));
        } else {
            var name = players.get(player);
            addMessage(game.messages.format("game.action."  + (changed ? "change" : "make") + "_prediction.other", name, prediction));
        }

        var index = orderedPlayers.indexOf(player);
        if (index == -1) throw new NoSuchElementException();

        execute(() -> padOfTruth.setPrediction(index, round, prediction));
    }

    /**
     * Removes the card from the players hand and puts it into the {@linkplain #cardStack stack}.
     */
    public void playCard(@NotNull UUID player, @NotNull Card card) {
        if (isSelf(player)) {
            addMessage(game.messages.get("game.action.play_card.self"));
        } else {
            var name = players.get(player);
            addMessage(game.messages.format("game.action.play_card.other", name));
        }

        Seat seat = seats.getOrDefault(player, Seat.FALLBACK);

        var sequence = sequence();
        sequence.addAction(run(() -> {
            CardActor actor = null;
            if (isSelf(player)) {
                actor = handCards.remove(card);
            }

            if (actor == null) {
                actor = seat.createHandCard(card, atlas);
            }

            cardStack.add(seat, actor);
            sequence.addAction(delay(actor));
            sequence.addAction(delay(AnimationTimings.STACK_HOLD));
        }));
        execute(sequence);
    }

    /**
     * Adds the scores for a round to the corresponding row of the {@linkplain #padOfTruth pad of truth}.
     */
    public void addScores(int round, @NotNull Map<UUID, Integer> scores) {
        execute(() -> {
            for (int i = 0; i < orderedPlayers.size(); i++) {
                UUID player = orderedPlayers.get(i);
                padOfTruth.setScore(i, round, scores.getOrDefault(player, 0));
            }
        });
    }

    /**
     * Removes all visual changes done by {@link #setActivePlayer(UUID, UserInputMessage.Action, long)}.
     */
    public void clearActivePlayer() {
        setActivePlayer(null, null, 0);
    }

    /**
     * Highlights the given players label and sets the {@linkplain #setPersistentMessage(String) persistent message}
     * accordingly.
     */
    public void setActivePlayer(@Nullable UUID player, @Nullable UserInputMessage.Action action, long timeout) {
        execute(() -> nameLabels.forEach((p, l) -> l.setStyle(p.equals(player) ? labelStyleActive : labelStyleDefault)));

        if (isSelf(player) || action == null) {
            setPersistentMessage(null);
        } else {
            var key = switch (action) {
                case CHANGE_PREDICTION -> "game.message.change_prediction.";
                case JUGGLE_CARD -> "game.message.juggle_card.";
                case MAKE_PREDICTION -> "game.message.make_prediction.";
                case PICK_TRUMP -> "game.message.pick_trump.";
                case PLAY_CARD -> "game.message.play_card.";
                default -> throw new AssertionError();
            };

            if (player != null) {
                setPersistentMessage(game.messages.format(key + "other", players.get(player)));
            } else {
                setPersistentMessage(game.messages.get(key + "all"));
            }
        }
    }

    /**
     * Visually highlights the given card on the players hand.
     */
    public void setSelectedCard(@Nullable Card card) {
        handCards.setSelected(card);
    }

    /**
     * Swaps the current {@linkplain #trumpCardActor trump card} (if present) with a {@linkplain Card#WEREWOLF werewolf}
     * card on the given players hand.
     */
    public void swapTrumpCard(@NotNull UUID player) {
        var seat = seats.getOrDefault(player, Seat.FALLBACK);

        var sequence = sequence();
        sequence.addAction(run(() -> {
            if (trumpCardActor == null || !trumpCardActor.hasParent()) return;

            if (isSelf(player)) {
                var handCard = handCards.find(Card.WEREWOLF);
                if (handCard == null) return;

                handCard.setCard(trumpCardActor.getCard());
                trumpCardActor.setCard(Card.WEREWOLF);

                float localHandX = handCard.getX(), localHandY = handCard.getY();
                float stageTrumpX = 10, stageTrumpY = 10;
                float localTrumpX, localTrumpY;
                float stageHandX, stageHandY;

                var pos = new Vector2(handCard.getX(), handCard.getY());
                handCards.localToStageCoordinates(pos);
                stageHandX = pos.x;
                stageHandY = pos.y;

                pos.set(stageTrumpX, stageTrumpY);
                handCards.stageToLocalCoordinates(pos);
                localTrumpX = pos.x;
                localTrumpY = pos.y;

                trumpCardActor.setPosition(stageHandX, stageHandY);
                trumpCardActor.setRotation(0);

                handCard.setPosition(localTrumpX, localTrumpY);
                handCard.setRotation(0);

                sequence.addAction(parallel(
                        targeting(trumpCardActor, moveTo(stageTrumpX, stageTrumpY, AnimationTimings.WEREWOLF_SWAP)),
                        targeting(handCard, moveTo(localHandX, localHandY, AnimationTimings.WEREWOLF_SWAP))
                ));
            } else {
                var handCard = new CardActor(trumpCardActor.getCard(), atlas);
                handCard.setPosition(10, 10);
                trumpCardActor.setCard(Card.WEREWOLF);
                game.data.stage.addActor(handCard);

                sequence.addAction(seat.moveToHand(trumpCardActor, 0));
                sequence.addAction(parallel(
                        targeting(trumpCardActor, moveTo(10,10, AnimationTimings.WEREWOLF_SWAP)),
                        targeting(trumpCardActor, rotateTo(0, AnimationTimings.WEREWOLF_SWAP)),
                        seat.moveToHand(handCard, AnimationTimings.WEREWOLF_SWAP)
                ));
                sequence.addAction(removeActor(handCard));
            }
        }));
        execute(sequence);
    }

    //<editor-fold desc="Overlays" defaultstate="collapsed">
    public void showTrumpOverlay(@Nullable UUID player, @NotNull Card trumpCard, @Nullable Card.Suit trumpSuit) {
        if (trumpCardActor == null) {
            trumpCardActor = new CardActor(Card.HIDDEN, atlas);
        }

        if (trumpSuitActor == null) {
            trumpSuitActor = new CardActor(Card.HIDDEN, atlas);
        }

        execute(new TrumpOverlay(this, players.get(player), trumpCard, trumpSuit));
    }

    public void showColoredCardOverlay(@NotNull Card card, long timeout) {
        if (card == Card.JUGGLER) {
            execute(new PlayColoredCardOverlay(this, timeout, card, Card.JUGGLER_RED, Card.JUGGLER_GREEN, Card.JUGGLER_BLUE, Card.JUGGLER_YELLOW));
        } else if (card == Card.CLOUD) {
            execute(new PlayColoredCardOverlay(this, timeout, card, Card.CLOUD_RED, Card.CLOUD_GREEN, Card.CLOUD_BLUE, Card.CLOUD_YELLOW));
        } else {
            throw new IllegalArgumentException();
        }
    }

    public InteractionOverlay showPickTrumpOverlay(long timeout, boolean allowNone) {
        var overlay = new PickTrumpOverlay(this, timeout, allowNone);
        execute(overlay);
        return overlay;
    }

    public void showScoreOverlay() {
        execute(sequence(
                run(() -> padOfTruth.setEnabled(false)),
                parallel(
                        targeting(padOfTruth, scaleTo(1, 1, AnimationTimings.OVERLAY_SHARED_ELEMENT)),
                        targeting(padOfTruth, moveTo((WizardGame.WIDTH - padOfTruth.getWidth()) / 2, (WizardGame.HEIGHT - padOfTruth.getHeight()) / 2, AnimationTimings.OVERLAY_SHARED_ELEMENT))
                ),
                delay(AnimationTimings.OVERLAY_HOLD),
                parallel(
                        targeting(padOfTruth, scaleTo(PadOfTruth.COLLAPSED_SCALE, PadOfTruth.COLLAPSED_SCALE, AnimationTimings.OVERLAY_SHARED_ELEMENT)),
                        targeting(padOfTruth, moveTo(WizardGame.WIDTH - 10 - PadOfTruth.EXTENDED_WIDTH, 10))
                ),
                run(() -> padOfTruth.setEnabled(true))
        ));
    }

    public InteractionOverlay showMakePredictionOverlay(int round, long timeout) {
        var overlay = new MakePredictionOverlay(this, timeout, round);
        execute(overlay);
        return overlay;
    }

    public InteractionOverlay showChangePredictionOverlay(int round, int oldPrediction, long timeout) {
        var overlay = new ChangePredictionOverlay(this, timeout, round, oldPrediction);
        execute(overlay);
        return overlay;
    }
    //</editor-fold>

    //<editor-fold desc="Messages" defaultstate="collapsed">
    public void addMessage(@Nls String text) {
        addMessage(text, false);
    }

    public void addMessage(@Nls String text, boolean immediate) {
        var label = new Label(text, game.data.skin);
        label.addAction(sequence(
                delay(AnimationTimings.MESSAGE_HOLD),
                alpha(0, AnimationTimings.MESSAGE_FADE),
                removeActor()
        ));

        Runnable runnable = () -> {
            if (persistentMessage != null && persistentMessage.getParent() == messages) {
                messages.addActorBefore(persistentMessage, label);
            } else {
                messages.addActor(label);
            }
        };

        if (immediate) {
            runnable.run();
        } else {
            execute(runnable);
        }
    }

    public void setPersistentMessage(@Nls String text) {
        execute(() -> {
            if (persistentMessage != null) {
                persistentMessage.addAction(sequence(
                        delay(AnimationTimings.MESSAGE_HOLD),
                        alpha(0, AnimationTimings.MESSAGE_FADE),
                        removeActor()
                ));
                persistentMessage = null;
            }

            if (text != null) {
                persistentMessage = new Label(text, game.data.skin);
                messages.addActor(persistentMessage);
            }
        });
    }
    //</editor-fold>

    public void ready(boolean success) {
        if (success) closeInteractionOverlay();
    }

    public void timeout() {
        addMessage(game.messages.get("game.message.timeout"));
        closeInteractionOverlay();
        clearActivePlayer();
    }

    public void sync() {
        pendingSync.set(true);
    }

    public void closeInteractionOverlay() {
        execute(() -> {
            if (currentAction instanceof Overlay overlay && overlay instanceof InteractionOverlay) {
                overlay.close();
            }
        });
    }

    @Override
    public void dispose() {
        if (atlas != null) atlas.dispose();
    }

    public WizardGame.Data getData() {
        return game.data;
    }

    public I18NBundle getMessages() {
        return game.messages;
    }

    private Action animateJuggle(@NotNull List<CardActor> removed, @NotNull List<CardActor> added) {
        // find left- and rightmost seat
        Seat tmpLeft = null, tmpRight = null;
        for (Seat seat : seats.values()) {
            if (seat == Seat.BOTTOM) continue;
            if (tmpLeft == null || tmpLeft.compareTo(seat) > 0) tmpLeft = seat;
            if (tmpRight == null || tmpRight.compareTo(seat) < 0) tmpRight = seat;
        }
        if (tmpLeft == null) return null;
        var left = tmpLeft;
        var right = tmpRight;

        var animation = parallel();
        removed.forEach(actor -> {
            game.data.stage.addActor(actor);
            animation.addAction(left.moveToHand(actor, AnimationTimings.JUGGLE));
        });

        added.forEach(actor -> {
            // capture old values taking consideration of the animated layout changes in CardsGroup
            float x = actor.getX(), y = actor.getY();
            var actions = actor.getActions();
            for (int i = 0; i < actions.size; i ++) {
                var action = actions.get(i);
                if (action instanceof MoveToAction move) {
                    x = move.getX();
                    y = move.getY();
                    actions.removeIndex(i--);
                }
            }
            var rotation = actor.getRotation();

            // apply start values
            var startPos = new Vector2(right.getHandX(), right.getHandY());
            handCards.stageToLocalCoordinates(startPos);
            animation.addAction(targeting(actor, moveTo(startPos.x, startPos.y)));
            animation.addAction(targeting(actor, rotateTo(right.getHandAngle())));

            // animate
            animation.addAction(targeting(actor, moveTo(x, y, AnimationTimings.JUGGLE)));
            animation.addAction(targeting(actor, rotateTo(rotation, AnimationTimings.JUGGLE)));
        });

        var cleanup = parallel();
        removed.forEach(actor -> cleanup.addAction(removeActor(actor)));

        return sequence(animation, cleanup);
    }

    @Getter
    public enum Seat {
        FALLBACK(WizardGame.WIDTH * 0.5f, WizardGame.HEIGHT * 0.5f, 0, 0, 0, WizardGame.WIDTH * 0.5f, WizardGame.HEIGHT * 0.5f),
        BOTTOM(WizardGame.WIDTH * 0.5f, 0, 0, 0, Align.bottom, WizardGame.WIDTH * 0.5f, 300),
        LEFT(0, WizardGame.HEIGHT * 0.5f, 50, WizardGame.HEIGHT * 0.5f + 110f, Align.bottomLeft, 117.5f, WizardGame.HEIGHT * 0.5f),
        TOP_LEFT(WizardGame.WIDTH * 0.25f, WizardGame.HEIGHT, WizardGame.WIDTH * 0.25f, WizardGame.HEIGHT - 50, Align.top, WizardGame.WIDTH * 0.25f, WizardGame.HEIGHT - 200),
        TOP(WizardGame.WIDTH * 0.5f, WizardGame.HEIGHT, WizardGame.WIDTH * 0.5f, WizardGame.HEIGHT - 50, Align.top, WizardGame.WIDTH * 0.5f, WizardGame.HEIGHT - 200),
        TOP_RIGHT(WizardGame.WIDTH * 0.75f, WizardGame.HEIGHT, WizardGame.WIDTH * 0.75f, WizardGame.HEIGHT - 50, Align.top, WizardGame.WIDTH * 0.75f, WizardGame.HEIGHT - 200),
        RIGHT(WizardGame.WIDTH, WizardGame.HEIGHT * 0.5f, WizardGame.WIDTH - 50, WizardGame.HEIGHT * 0.5f + 110f, Align.bottomRight, WizardGame.WIDTH - 117.5f, WizardGame.HEIGHT * 0.5f);

        // position of the hand, should be offscreen
        private final float handX;
        private final float handY;
        private final float handAngle;

        // position of the label
        private final float labelX;
        private final float labelY;
        private final int labelAlign;

        private final float frontX;
        private final float frontY;

        Seat(float handX, float handY, float labelX, float labelY, int labelAlign, float frontX, float frontY) {
            this.handX = handX;
            this.handY = handY;
            this.labelX = labelX;
            this.labelY = labelY;
            this.labelAlign = labelAlign;
            this.frontX = frontX;
            this.frontY = frontY;

            var deltaX = WizardGame.WIDTH * 0.5f - handX;
            var deltaY = WizardGame.HEIGHT * 0.5f - handY;
            this.handAngle = (float) Math.toDegrees(Math.atan2(deltaY, deltaX) + Math.PI / 2);
        }

        public CardActor createHandCard(Card card, TextureAtlas atlas) {
            var actor = new CardActor(card, atlas);
            actor.setPosition(getHandX() - actor.getWidth() / 2, getHandY() - actor.getHeight() / 2);
            actor.setRotation(getHandAngle());
            return actor;
        }

        public Action moveToHand(CardActor actor, float duration) {
            return parallel(
                    targeting(actor, moveTo(handX, handY, duration)),
                    targeting(actor, rotateTo(handAngle, duration))
            );
        }
    }
}
